/*****************************************************************************/
/*                      WiNRADiO G35DDC EXCALIBUR ULTRA                      */
/*                                SDK DEMO                                   */
/*                   Interlayer between G35DDC API and GUI                   */
/*                        (C) Copyright RADIXON 2014                         */
/*****************************************************************************/

#include "u_receiver.h"
#include "u_fft.h"
#include <math.h>

extern G3XDDCAPI_CREATE_INSTANCE CreateInstance;

const wxChar* TReceiver::m_ModeNames[DEM_MODE_COUNT]=
{
    wxT("AM"),   //G3XDDC_MODE_AM
    wxT("AMS"),  //G3XDDC_MODE_AMS
    wxT("LSB"),  //G3XDDC_MODE_LSB
    wxT("USB"),  //G3XDDC_MODE_USB
    wxT("DSB"),  //G3XDDC_MODE_DSB
    wxT("ISB"),  //G3XDDC_MODE_ISB
    wxT("CW"),   //G3XDDC_MODE_CW
    wxT("FM"),   //G3XDDC_MODE_FM
};

uint32_t TReceiver::m_ModeCodes[DEM_MODE_COUNT]=
{
    G3XDDC_MODE_AM,
    G3XDDC_MODE_AMS,
    G3XDDC_MODE_LSB,
    G3XDDC_MODE_USB,
    G3XDDC_MODE_DSB,
    G3XDDC_MODE_ISB,
    G3XDDC_MODE_CW,
    G3XDDC_MODE_FM
};

int TReceiver::m_Presel_CutOffLow[PRESEL_COUNT]=
{
    0,850000,2400000,5400000,11800000
};

int TReceiver::m_Presel_CutOffHigh[PRESEL_COUNT]=
{
    3100000,5400000,11800000,23300000,50000000
};

//TReceiver::TReceiver
//Constructor
TReceiver::TReceiver(void):m_IF_FFT_Mutex(wxMUTEX_RECURSIVE),m_DDC1_FFT_Mutex(wxMUTEX_RECURSIVE),
    m_DDC1_FFT_OverlappingBuffer(MAX_FFT_SIZE)
{ 
    m_Device=NULL;
    m_Callback=NULL;   
    m_Power=false;
    m_Ignore=1; //disable all Set commands while receiver is not open
    m_Channel=NULL;
}

//TReceiver::~TReceiver
//Destructor
TReceiver::~TReceiver(void)
{
    Close();
}


//TReceiver::InvokeEvent 
//Passes generated receiver event to GUI via callback 
void TReceiver::InvokeEvent(TReceiverEvent Event)
{
    if(m_Callback)
    {     
        //ignore all Set commands (e.g. SetPower, SetDDC1Frequency,...) during event
        m_Ignore++;

        m_Callback->ReceiverCallback_Event(Event);

        m_Ignore--;
    }
}

//TReceiver::InvokeEvent 
//Passes generated channel event to GUI via callback
void TReceiver::InvokeEvent(uint32_t Channel,TChannelEvent Event)
{
    if(m_Callback)
    {     
        //ignore all Set... commands
        m_Ignore++;

        m_Callback->ReceiverCallback_ChannelEvent(Channel,Event);

        m_Ignore--;
    }
}

//TReceiver::Open 
//Opens G35DDC device and initializes object members
bool TReceiver::Open(const char *DevicePath)
{
 uint32_t i; 

    Close();

    //create instance of G35DDC device object calling G35DDC API function
    if(!CreateInstance(G35DDC_CLASS_ID_DEVICE,(void**)&m_Device))
    {
        return false;
    }

    if(!m_Device->Open(DevicePath))
    {
        return false;
    }
    

    m_Ignore--; //allow receiver and channel events

    m_Device->GetDeviceInfo(&m_DeviceInfo,sizeof(m_DeviceInfo));

    m_Channel=new TChannel[m_DeviceInfo.ChannelCount];

    m_Power=false; 
    

    m_Attenuator=ATTENUATOR_0dB;
    m_Device->SetAttenuator(ATTENUATOR_0dB*3);

    m_Preamp=false;
    m_Device->SetPreamp(FALSE);
   
    m_Presel_CutOffLowIndex=0;
    m_Presel_CutOffHighIndex=PRESEL_COUNT-1;
    m_Device->SetPreselectors(m_Presel_CutOffLow[m_Presel_CutOffLowIndex],m_Presel_CutOffHigh[m_Presel_CutOffHighIndex]);

    m_ADCLevel=0;
    m_ActiveChannel=0;
    

    m_IF_RBW_Index=3;
    UpdateIFFFTCoeffs();


    //DDC1
    m_Device->GetDDC1(&m_DDC1_Index,&m_DDC1_Info);
    m_Device->GetDDC1Count(&m_DDC1_Count);
    m_Device->GetDDC1Frequency((uint32_t*)&m_DDC1_Frequency);
    m_DDC1_RBW_Index=0;
    UpdateDDC1FFTCoeffs();

    //DDC2
    m_Device->GetDDC2(&m_DDC2_Index,&m_DDC2_Info);

    for(i=0;i<m_DeviceInfo.ChannelCount;i++)
    {
        //DDC2
        m_Device->GetDDC2Frequency(i,&m_Channel[i].m_DDC2_Frequency);
        m_Channel[i].m_DDC2_AbsFrequency=m_Channel[i].m_DDC2_Frequency+m_DDC1_Frequency;
        m_Channel[i].m_DDC2_RBW_Index=0;
        UpdateDDC2FFTCoeffs(i);

        //DEM
        m_Device->GetDemodulatorFrequency(i,&m_Channel[i].m_DEM_Frequency);
        m_Channel[i].m_DEM_AbsFrequency=m_Channel[i].m_DEM_Frequency+m_Channel[i].m_DDC2_AbsFrequency;
        m_Device->GetDemodulatorFilterBandwidth(i,&m_Channel[i].m_DEM_Bandwidth);
        
        m_Channel[i].m_DEM_Mode=0;
        m_Device->SetDemodulatorMode(i,m_ModeCodes[0]);

        m_Channel[i].m_DEM_AMSSideband=G3XDDC_SIDE_BAND_BOTH;
        m_Device->SetDemodulatorParam(i,G3XDDC_DEMODULATOR_PARAM_AMS_SIDE_BAND,&m_Channel[i].m_DEM_AMSSideband,sizeof(m_Channel[i].m_DEM_AMSSideband));

        m_Channel[i].m_DEM_DSBSideband=G3XDDC_SIDE_BAND_BOTH;
        m_Device->SetDemodulatorParam(i,G3XDDC_DEMODULATOR_PARAM_DSB_SIDE_BAND,&m_Channel[i].m_DEM_DSBSideband,sizeof(m_Channel[i].m_DEM_DSBSideband));

        m_Channel[i].m_DEM_ISBSideband=G3XDDC_SIDE_BAND_BOTH;
        m_Device->SetDemodulatorParam(i,G3XDDC_DEMODULATOR_PARAM_ISB_SIDE_BAND,&m_Channel[i].m_DEM_ISBSideband,sizeof(m_Channel[i].m_DEM_ISBSideband));

        m_Channel[i].m_DEM_CWTone=800;
        m_Device->SetDemodulatorParam(i,G3XDDC_DEMODULATOR_PARAM_CW_FREQUENCY,&m_Channel[i].m_DEM_CWTone,sizeof(m_Channel[i].m_DEM_CWTone));

        //Audio        
        m_Channel[i].m_Audio_RBW_Index=0;
        UpdateAudioFFTCoeffs(i);
       

        m_Channel[i].m_Audio_Filter=false;
        m_Channel[i].m_Audio_Filter_CutOffLow=300;
        m_Channel[i].m_Audio_Filter_CutOffHigh=3000;
        m_Channel[i].m_Audio_Filter_Deemphasis=0;

        m_Device->SetAudioFilter(i,FALSE);
        m_Device->SetAudioFilterParams(i,300,3000,0);

        m_Channel[i].m_AGC=AGC_SLOW;
        UpdateAGC(i);

        m_Channel[i].m_Gain=0;
        m_Device->SetGain(i,0);

        m_Channel[i].m_Audio_Volume=20;
        UpdateAudioVolumeCoeff(i);

        if(m_ActiveChannel==i)
        {
            m_Channel[i].m_Audio_MuteBits=0;
        }
        else
        {
            m_Channel[i].m_Audio_MuteBits=MUTE_BITS_INACTIVE_CHANNEL;
        }

        m_Device->SetMute(i,m_Channel[i].m_Audio_MuteBits!=0);
    }

    //set callback for API G35DDC device object to this object
    m_Device->SetCallback(this);

    return true;
}

//TReceiver::Close
//Closes G35DDC device if open
void TReceiver::Close(void)
{
    if(m_Device==NULL)
    {
        return;
    }

    m_Ignore++;
    
    //turn receiver off, it stops all streamings too
    m_Device->SetPower(FALSE);
    m_Device->Release();
    m_Device=NULL;
    
    delete[] m_Channel;
    m_Channel=NULL;

    m_Power=false;
}

//TReceiver::SetPower
//Turns receiver on/off
void TReceiver::SetPower(bool Power)
{
    if(m_Ignore)
    {
        return;
    }

    if(m_Power!=Power)
    {
        if(!Power)
        {            
            StopDDC1();
            m_Device->StopIF();
        }

        if(m_Device->SetPower(Power))
        {
            m_Power=Power;

            OnPowerChanged();
        }
    }
}

//TReceiver::OnPowerChanged
//Starts streamings if receiver is turned on and invokes power change event for GUI
void TReceiver::OnPowerChanged(void)
{
 uint32_t i;

    if(m_Power)
    {
        //start IF snapshots
        m_Device->StartIF(UPDATE_INTERVAL);

        //start DDC1
        StartDDC1();

        //start DDC2 and audio streaming for all the channels
        for(i=0;i<m_DeviceInfo.ChannelCount;i++)
        {            
            StartDDC2(i);
            StartAudio(i);
        }        
    }
    else
    {
        m_ADCLevel=0;
        //it is not required to stop IF, DDC1, DDC2 and audio streaming here, all is stopped automatically when power off
    }

    InvokeEvent(ReceiverEvent_Power);    
}

//TReceiver::SetAttenuator
//Sets HF attenuator
void TReceiver::SetAttenuator(uint32_t Attenuator)
{
    if(m_Ignore)
    {
        return;
    }

    if(Attenuator>ATTENUATOR_MAX)
    {
        Attenuator=ATTENUATOR_MAX;
    }

    if(m_Attenuator!=Attenuator)
    {
        m_Attenuator=Attenuator;
        OnAttenuatorChanged();
    }
}

//TReceiver::OnAttenuatorChanged
//Updates HW settings when HF attenuator is changed and invokes event for GUI
void TReceiver::OnAttenuatorChanged(void)
{
    m_Device->SetAttenuator(m_Attenuator*3);

    UpdateSpectrumCompensationData();

    InvokeEvent(ReceiverEvent_Attenuator);
}

//TReceiver::SetPreamplifier
//Sets VHF/UHF preamplifier
void TReceiver::SetPreamplifier(bool Preamp)
{
    if(m_Ignore)
    {
        return;
    }

    if(m_Preamp!=Preamp)
    {
        m_Preamp=Preamp;

        OnPreaplifierChanged();
    }
}

//TReceiver::OnPreamplifierChanged
//Updates HW settings when VHF/UHF preamplifier is changed and invokes event for GUI
void TReceiver::OnPreaplifierChanged(void)
{
    m_Device->SetPreamp(m_Preamp);

    UpdateSpectrumCompensationData();

    InvokeEvent(ReceiverEvent_Preamplifier);    
}

//TReceiver::SetDDC1
//Changes DDC1 BW for specified channel and restarts streaming
void TReceiver::SetDDC1(uint32_t TypeIndex)
{
 uint32_t i;

    if(m_Ignore)
    {
        return;
    }


    if(TypeIndex>=m_DDC1_Count)
    {
        TypeIndex=m_DDC1_Count-1;
    }

    if(m_DDC1_Index!=TypeIndex)
    {
        StopDDC1();

        m_DDC1_Index=TypeIndex;

        OnDDC1Changed();

        StartDDC1();

        for(i=0;i<m_DeviceInfo.ChannelCount;i++)
        {            
            StartDDC2(i);
            StartAudio(i);
        }
    }
}

//TReceiver::OnDDC1Changed
//Updates HW settings when DDC1 BW is changed, invokes events for GUI
void TReceiver::OnDDC1Changed(void)
{
 uint32_t DDC2TypeIndex,i;
 int32_t Min,Max;
 
 
    //change HW settings
    m_Device->SetDDC1(m_DDC1_Index);
    //get info about new DDC type
    m_Device->GetDDCInfo(m_DDC1_Index,&m_DDC1_Info);

    UpdateDDC1FFTCoeffs();

    InvokeEvent(ReceiverEvent_DDC1);
    InvokeEvent(ReceiverEvent_DDC1_RBWList);

    //the following ensures entire DDC1 will fall to IF window 
    //for example, if previous DDC1 BW was 20kHz, and it was on the left side of the IF window
    //(10 kHz) and new DDC1 BW is 4MHz, the DDC1 relative frequency has to be
    //modified to 2MHz, to have entire DDC1 in IF window on left side of the
    //IF window
    
    Min=m_DDC1_Info.Bandwidth/2;    
    Max=50000000-m_DDC1_Info.Bandwidth/2;

    if(m_DDC1_Frequency<Min)
    {
        m_DDC1_Frequency=Min;
        OnDDC1FrequencyChanged(true);
    }
    else if(m_DDC1_Frequency>Max)
    {
        m_DDC1_Frequency=Max;
        OnDDC1FrequencyChanged(true);
    }

    m_Device->GetDDC2(&DDC2TypeIndex,NULL);



    if(m_DDC2_Index!=DDC2TypeIndex)
    {
        m_DDC2_Index=DDC2TypeIndex;
        OnDDC2Changed();
    }
    else
    {
        //the following ensures entire DDC2 will fall to DDC1
        Max=(m_DDC1_Info.Bandwidth-m_DDC2_Info.Bandwidth)/2;
        Min=-Max;

        for(i=0;i<m_DeviceInfo.ChannelCount;i++)
        {
            if(m_Channel[i].m_DDC2_Frequency<Min)
            {
                m_Channel[i].m_DDC2_Frequency=Min;

                OnDDC2FrequencyChanged(i);
            }
            else if(m_Channel[i].m_DDC2_Frequency>Max)
            {
                m_Channel[i].m_DDC2_Frequency=Max;

                OnDDC2FrequencyChanged(i);
            }
        }
    }    
}



//TReceiver::OnDDC2Changed
//Updates HW settings when DDC2 BW is changed and invokes events for GUI
void TReceiver::OnDDC2Changed(void)
{
 int32_t Min,Max;
 int32_t DemMin,DemMax;
 uint32_t i;
        
    //get info (bw and sample rate) about new DDC type of DDC2
    m_Device->GetDDCInfo(m_DDC2_Index,&m_DDC2_Info);

    InvokeEvent(ReceiverEvent_DDC2);
    InvokeEvent(ReceiverEvent_DDC2_RBWList);
    
    
    Max=(m_DDC1_Info.Bandwidth-m_DDC2_Info.Bandwidth)/2;
    Min=-Max;

    for(i=0;i<m_DeviceInfo.ChannelCount;i++)
    {
        UpdateDDC2FFTCoeffs(i);

        //the following ensures entire DDC2 will fall to DDC1
        
        if(m_Channel[i].m_DDC2_Frequency>Max)
        {
            m_Channel[i].m_DDC2_Frequency=Max;

            OnDDC2FrequencyChanged(i);
        }
        else if(m_Channel[i].m_DDC2_Frequency<Min)
        {
            m_Channel[i].m_DDC2_Frequency=Min;

            OnDDC2FrequencyChanged(i);
        }
    

        if(m_Channel[i].m_DEM_Bandwidth>m_DDC2_Info.Bandwidth)
        {
            //new DDC2 BW does not allow current DEM BW, it has to be reduced
            m_Channel[i].m_DEM_Bandwidth=m_DDC2_Info.Bandwidth;

            OnDemBandwidthChanged(i);
        }
        else
        {
            //the following ensures current demod filter will fall to DDC2
            GetDemFrequencyMinMax(i,&DemMin,&DemMax);

            if(m_Channel[i].m_DEM_Frequency<DemMin)
            {
                m_Channel[i].m_DEM_Frequency=DemMin;

                OnDemFrequencyChanged(i);
            }
            else if(m_Channel[i].m_DEM_Frequency>DemMax)
            {
                m_Channel[i].m_DEM_Frequency=DemMax;

                OnDemFrequencyChanged(i);
            }
        }        
    }
}


//TReceiver::G35DDC_DDC1StreamCallback
//Implementation of IG35DDCDeviceCallback interface method
//Passes DDC1 samples to recorder (WavWriter), calculates FFT, applies
//spectrum compensation and passes final frequency spectrum to GUI via callback
void G35DDC_CALLBACK TReceiver::G35DDC_DDC1StreamCallback(IG35DDCDevice *Device,const void *Buffer,uint32_t NumberOfSamples,uint32_t BitsPerSample)
{
 uint32_t i,j,BinCount,FFTSize;
 float m;
 float *FFTBuffer,TempI,TempQ,*FFTCoeffs;
 float *Comp;
 const float *OverlBuffer;

    m_DDC1_FFT_Mutex.Lock();

    //write samples to overlapping FFT buffer
    if(BitsPerSample==16)
    {
        m_DDC1_FFT_OverlappingBuffer.AddSamples_16((const int16_t*)Buffer,NumberOfSamples);
    }
    else
    {
        m_DDC1_FFT_OverlappingBuffer.AddSamples_32((const int32_t*)Buffer,NumberOfSamples);
    }

    OverlBuffer=m_DDC1_FFT_OverlappingBuffer.Get();

    if(OverlBuffer!=NULL)
    {
        FFTSize=m_DDC1_FFT_Size;
        FFTBuffer=m_DDC1_FFT_Buffer;
        FFTCoeffs=m_DDC1_FFT_Coeffs;

        //apply FFT normalization and window coefficients at once
        for(i=0;i<FFTSize;i++)
        {
            FFTBuffer[i*2]=OverlBuffer[i*2]*FFTCoeffs[i];
            FFTBuffer[i*2+1]=OverlBuffer[i*2+1]*FFTCoeffs[i];
        }

        //FFT
        FFT(FFTBuffer,FFTSize);

        //swap lower and upper half of the FFT result
        BinCount=FFTSize/2;
        for(i=0;i<BinCount;i++)
        {
            TempI=FFTBuffer[i*2];
            TempQ=FFTBuffer[i*2+1];

            FFTBuffer[i*2]=FFTBuffer[(i+BinCount)*2];
            FFTBuffer[i*2+1]=FFTBuffer[(i+BinCount)*2+1];

            FFTBuffer[(i+BinCount)*2]=TempI;
            FFTBuffer[(i+BinCount)*2+1]=TempQ;
        }

        //now frequency spectrum after this FFT looks like the following
        // |                                     |                                     |
        // |                                     |                                     |
        // |------*******************************|*******************************------|
        // -DDC1 sample rate/2        baseband frequency (0)        +DDC1 sample rate /2
        // The first FFT bin               Middle FFT bin               The last FFT bin
    
        //asterisks show useful band (G3XDDC_DDC_INFO.Bandwidth) in FFT result

        //the following converts amplitudes to dB and applies compensation coeffs
        //to have absolute values in dBm instead of relative ones in dB
        BinCount=m_DDC1_FFT_BinCount;
        Comp=m_DDC1_SpectrumCompensation;

        for(i=0,j=m_DDC1_FFT_FirstBin;i<BinCount;i++,j++)
        {
            m=FFTBuffer[j*2]*FFTBuffer[j*2]+FFTBuffer[j*2+1]*FFTBuffer[j*2+1];

            if(m>0)
            {
                FFTBuffer[j]=10*log10(m);
            }
            else
            {
                FFTBuffer[j]=-180;
            }

            FFTBuffer[j]+=Comp[j];
        }

        //send final DDC1 frequency spectrum to GUI
        m_Callback->ReceiverCallback_DDC1Spectrum(FFTBuffer+m_DDC1_FFT_FirstBin,
            BinCount,m_DDC1_FFT_Range,m_DDC1_FFT_Offset);
    }    

    m_DDC1_FFT_Mutex.Unlock();
}

//TReceiver::G35DDC_DDC1PlaybackStreamCallback
//Implementation of IG35DDCDeviceCallback interface method
//It does nothing in this demo.
BOOL G35DDC_CALLBACK TReceiver::G35DDC_DDC1PlaybackStreamCallback(IG35DDCDevice *Device,void *Buffer,uint32_t NumberOfSamples,uint32_t BitsPerSample)
{
    return FALSE;
}

//TReceiver::G35DDC_DDC2StreamCallback
//Implementation of IG35DDCDeviceCallback interface method
//Calculates FFT from DDC2 signal, applies spectrum compensation 
//and passes final frequency spectrum to GUI via callback
void G35DDC_CALLBACK TReceiver::G35DDC_DDC2StreamCallback(IG35DDCDevice *Device,uint32_t Channel,const float *Buffer,uint32_t NumberOfSamples)
{
 uint32_t i,j,BinCount,FFTSize;
 float *FFTBuffer,TempI,TempQ,m;
 float *Comp,*FFTCoeffs;
 const float *OverlBuffer;

    m_Channel[Channel].m_DDC2_FFT_Mutex.Lock();

    //write samples to overlapping FFT buffer
    if(m_Channel[Channel].m_DDC2_FFT_OverlappingBuffer.AddSamples_F32(Buffer,NumberOfSamples))
    {
        //overlapping buffer is complete, FFT can be done

        OverlBuffer=m_Channel[Channel].m_DDC2_FFT_OverlappingBuffer.Get();
        FFTBuffer=m_Channel[Channel].m_DDC2_FFT_Buffer;
        FFTCoeffs=m_Channel[Channel].m_DDC2_FFT_Coeffs;
        FFTSize=m_Channel[Channel].m_DDC2_FFT_Size;
        
        //apply normalization and window coefficients at once
        for(i=0;i<FFTSize;i++)
        {
            FFTBuffer[i*2]=OverlBuffer[i*2]*FFTCoeffs[i];
            FFTBuffer[i*2+1]=OverlBuffer[i*2+1]*FFTCoeffs[i];
        }

        //FFT
        FFT(FFTBuffer,FFTSize);        

        //swap lower and upper half of the FFT result
        BinCount=m_Channel[Channel].m_DDC2_FFT_Size/2;
        for(i=0;i<BinCount;i++)
        {
            TempI=FFTBuffer[i*2];
            TempQ=FFTBuffer[i*2+1];

            FFTBuffer[i*2]=FFTBuffer[(i+BinCount)*2];
            FFTBuffer[i*2+1]=FFTBuffer[(i+BinCount)*2+1];

            FFTBuffer[(i+BinCount)*2]=TempI;
            FFTBuffer[(i+BinCount)*2+1]=TempQ;
        }

        //now frequency spectrum after this FFT looks like the following
        // |                                     |                                     |
        // |                                     |                                     |
        // |------*******************************|*******************************------|
        // -DDC2 sample rate/2        baseband frequency (0)        +DDC2 sample rate /2
        // The first FFT bin               Middle FFT bin               The last FFT bin
    
        //asterisks show useful band (G3XDDC_DDC_INFO.Bandwidth) in FFT result

        //the following converts amplitudes to dB and applies compensation coeffs
        //to have absolute values in dBm instead of relative ones in dB
        BinCount=m_Channel[Channel].m_DDC2_FFT_BinCount;   
        Comp=m_Channel[Channel].m_DDC2_SpectrumCompensation;

        for(i=0,j=m_Channel[Channel].m_DDC2_FFT_FirstBin;i<BinCount;i++,j++)
        {
            m=FFTBuffer[j*2]*FFTBuffer[j*2]+FFTBuffer[j*2+1]*FFTBuffer[j*2+1];

            if(m>0)
            {
                FFTBuffer[j]=10*log10(m);
            }
            else
            {
                FFTBuffer[j]=-180;
            }

            FFTBuffer[j]+=Comp[j];
        }

        //send final DDC2 frequency spectrum to GUI
        m_Callback->ReceiverCallback_DDC2Spectrum(Channel,
            FFTBuffer+m_Channel[Channel].m_DDC2_FFT_FirstBin,
            BinCount,
            m_Channel[Channel].m_DDC2_FFT_Range,m_Channel[Channel].m_DDC2_FFT_Offset);
    }    

    m_Channel[Channel].m_DDC2_FFT_Mutex.Unlock();
}

//TReceiver::G35DDC_DDC2PreprocessedStreamCallback
//Implementation of IG35DDCDeviceCallback interface method
void G35DDC_CALLBACK TReceiver::G35DDC_DDC2PreprocessedStreamCallback(IG35DDCDevice *Device,uint32_t Channel,const float *Buffer,uint32_t NumberOfSamples,float SlevelPeak,float SlevelRMS)
{
    //does nothing in this demo
}

//TReceiver::G35DDC_IFCallback
//Implementation of IG35DDCDeviceCallback interface method
//Calculates FFT from IF snapshot (snapshot directly from 100MHz ADC), applies spectrum compensation
//and pass final spectrum to GUI
void G35DDC_CALLBACK TReceiver::G35DDC_IFCallback(IG35DDCDevice *Device,CONST SHORT *Buffer,UINT32 NumberOfSamples,WORD MaxADCAmplitude,UINT32 ADCSamplingRate)
{
 uint32_t i;
 uint32_t Half;
 float m;

 
    m_ADCLevel=MaxADCAmplitude;

    m_IF_FFT_Mutex.Lock();

    //the following converts signed 16bit real signal to float (-1.0..+1.0)
    //and applies normalization and window coefficients at once.
    //Imaginary component is set 0 zero because input signal is real and input of used FFT 
    //is complex (I&Q) signal. It exists FFT algorithm optiomized for real signal which
    //does not require imaginary component at input, but it is not used in this demo
    for(i=0;i<m_IF_FFT_Size;i++)
    {
        m_IF_FFT_Buffer[i*2]=Buffer[i]*m_IF_FFT_Coeffs[i];
        m_IF_FFT_Buffer[i*2+1]=0;
    }

    //FFT
    FFT(m_IF_FFT_Buffer,m_IF_FFT_Size);
    
    //frequency spectrum after this FFT (with real signal at the input) looks like the following
    // |                                     |                                     |
    // |                                     |             inverted part           |
    // |-------------------------------------|-------------------------------------|
    // 0 Hz                               50 MHz                                ~0 Hz
    // The first FFT bin               Middle FFT bin               The last FFT bin        
    
    
    //the following converts amplitudes to dB and applies compensation coeffs
    //to have absolute values in dBm instead of relative ones in dB, this is done
    //only on useful band, rest of the spectrum is not used
    Half=m_IF_FFT_Size/2;        

    for(i=0;i<Half;i++)
    {
        m=m_IF_FFT_Buffer[i*2]*m_IF_FFT_Buffer[i*2]+m_IF_FFT_Buffer[i*2+1]*m_IF_FFT_Buffer[i*2+1];

        if(m>0)
        {
            m_IF_FFT_Buffer[i]=10*log10(m);
        }
        else
        {
            m_IF_FFT_Buffer[i]=-180;
        }

        m_IF_FFT_Buffer[i]+=m_IF_SpectrumCompensation[i];
    }

    //send final IF frequency spectrum to GUI
    m_Callback->ReceiverCallback_IFSpectrum(m_IF_FFT_Buffer,Half,50000000,0);

    m_IF_FFT_Mutex.Unlock();
}

//TReceiver::G35DDC_AudioStreamCallback
//Implementation of IG35DDCDeviceCallback interface method
//Calculates FFT from audio signal and pass final spectrum to GUI
void G35DDC_CALLBACK TReceiver::G35DDC_AudioStreamCallback(IG35DDCDevice *Device,UINT32 Channel,UINT32 Type,CONST FLOAT *Buffer,UINT32 NumberOfSamples)
{
 uint32_t i,Half,FFTSize;
 float *FFTBuffer;
 float *FFTCoeffs;
 const float *OverlBuffer;
 float m;
    
               

    switch(Type)
    {
        case 1:
            m_Channel[Channel].m_Audio_FFT_Mutex.Lock();
            
            if(m_Channel[Channel].m_Audio_FFT_OverlappingBuffer.AddSamples_Real_Mono_F32(Buffer,NumberOfSamples))
            {
                OverlBuffer=m_Channel[Channel].m_Audio_FFT_OverlappingBuffer.Get();
                FFTBuffer=m_Channel[Channel].m_Audio_FFT_Buffer;
                FFTCoeffs=m_Channel[Channel].m_Audio_FFT_Coeffs;
                FFTSize=m_Channel[Channel].m_Audio_FFT_Size;
                
                //convet to mode and apply normalization and window coeffs
                for(i=0;i<FFTSize;i++) 
                {
                    FFTBuffer[i*2]=OverlBuffer[i]*FFTCoeffs[i];
                    FFTBuffer[i*2+1]=0; //imaginary component is zero, audio signal does not have it
                }

                //FFT
                FFT(FFTBuffer,FFTSize);

                //frequency spectrum after this FFT (with real signal at the input) looks like the following
                // |                                     |                                     |
                // |                                     |             inverted part           |
                // |-------------------------------------|-------------------------------------|
                // 0 Hz                         audio_sample_rate/2                        ~0 Hz
                // The first FFT bin               Middle FFT bin               The last FFT bin 
                

                //the following converts amplitudes to dB on lower half of the FFT result,
                //inverted part is not used
                Half=FFTSize/2;
                
                for(i=0;i<Half;i++)
                {
                    m=FFTBuffer[i*2]*FFTBuffer[i*2]+FFTBuffer[i*2+1]*FFTBuffer[i*2+1];

                    if(m>0)
                    {
                        FFTBuffer[i]=10*log10(m);
                    }
                    else
                    {
                        FFTBuffer[i]=-180;
                    }
                }

                //send final audio frequency spectrum to GUI
                m_Callback->ReceiverCallback_AudioSpectrum(Channel,FFTBuffer,Half,AUDIO_SAMPLE_RATE/2);
            }   

            m_Channel[Channel].m_Audio_FFT_Mutex.Unlock();
            break;
    
        case 2:
            //send audio signal to GUI (to be played on waveout device - sound card)
            m_Callback->ReceiverCallback_Audio(Channel,Buffer,NumberOfSamples);
            break;

        default:
            break;
    }
}

//TReceiver::G35DDC_AudioPlaybackStreamCallback
//Implementation of IG35DDCDeviceCallback interface method
//It does nothing in this demo.
BOOL G35DDC_CALLBACK TReceiver::G35DDC_AudioPlaybackStreamCallback(IG35DDCDevice *Device,uint32_t ChannelIndex,float *Buffer,uint32_t NumberOfSamples)
{
    return FALSE;
}

//TReceiver::OnIFRBWChanged
//Updates FFT parameters and sends event to GUI when IF RBW is changed
void TReceiver::OnIFRBWChanged(void)
{
    UpdateIFFFTCoeffs();
    InvokeEvent(ReceiverEvent_IFRBWIndex);
}

//TReceiver::UpdateIFFFTCoeffs
//Updates IF spectrum FFT size acording to RBW index, normalization/window coeffs
//and IF spectrum compensation data
void TReceiver::UpdateIFFFTCoeffs(void)
{
    m_IF_FFT_Mutex.Lock();

    m_IF_FFT_Size=MIN_IF_FFT_SIZE<<m_IF_RBW_Index;

    GetNormalizedWindowCoeffs(m_IF_FFT_Coeffs,m_IF_FFT_Size,(1.0/32768));

    UpdateIFSpectrumCompensationData();

    m_IF_FFT_Mutex.Unlock();
}

//TReceiver::SetIFRBWIndex
//Changes IF RBW
void TReceiver::SetIFRBWIndex(uint32_t Index)
{
    if(m_Ignore)
    {
        return;
    }

    if(m_IF_RBW_Index!=Index)
    {
        m_IF_RBW_Index=Index;

        OnIFRBWChanged();
    }
}

//TReceiver::SetIFRBWIndex
//Returns available IF RBWs in Hz
void TReceiver::GetIFRBWs(double *RBWs)
{
 int i;

    for(i=0;i<RBW_COUNT;i++)
    {
        RBWs[i]=100000000.0/(MIN_IF_FFT_SIZE<<i);
    }
}

//TReceiver::UpdateDDC1FFTCoeffs
//Updates DDC1 spectrum FFT size acording to RBW index, normalization/window coeffs
//and DDC1 spectrum compensation data for given channel.
//It also computes position of useful band (G3XDDC_DDC_INFO.Bandwidth) in FFT result  
//(G3XDDC_DDC_INFO.SampleRate) 
void TReceiver::UpdateDDC1FFTCoeffs(void)
{
 double k,fFirst,fLast;
 int First,Last;

    m_DDC1_FFT_Mutex.Lock();    

    m_DDC1_FFT_Size=MIN_FFT_SIZE<<m_DDC1_RBW_Index;

    m_DDC1_FFT_OverlappingBuffer.SetSize(m_DDC1_FFT_Size);

    GetNormalizedWindowCoeffs(m_DDC1_FFT_Coeffs,m_DDC1_FFT_Size,1);

    k=(double)m_DDC1_Info.SampleRate/m_DDC1_FFT_Size;

    fFirst=(m_DDC1_Info.SampleRate-m_DDC1_Info.Bandwidth)*0.5/k;                           
    fLast=(m_DDC1_Info.SampleRate-(m_DDC1_Info.SampleRate-m_DDC1_Info.Bandwidth)*0.5)/k;

    First=(int)fFirst;          //the first FFT bin where useful band starts
    Last=(int)ceil(fLast);      //the last FFT bin where useful band ends

    m_DDC1_FFT_BinCount=Last-First+1;
    m_DDC1_FFT_FirstBin=First;

    m_DDC1_FFT_Range=m_DDC1_FFT_BinCount*k;
    m_DDC1_FFT_Offset=(fFirst-(First+0.5))*k;

    UpdateDDC1SpectrumCompensationData();

    m_DDC1_FFT_Mutex.Unlock();
}

//TReceiver::UpdateDDC2FFTCoeffs
//Updates DDC2 spectrum FFT size acording to RBW index, normalization/window coeffs
//and DDC1 spectrum compensation data for given channel.
//It also computes position of useful band (G3XDDC_DDC_INFO.Bandwidth) in FFT result  
//(G3XDDC_DDC_INFO.SampleRate) 
void TReceiver::UpdateDDC2FFTCoeffs(uint32_t Channel)
{
 double k,fFirst,fLast;
 int First,Last;
 TChannel &c=m_Channel[Channel];

    c.m_DDC2_FFT_Mutex.Lock();

    c.m_DDC2_FFT_Size=MIN_FFT_SIZE<<c.m_DDC2_RBW_Index;

    c.m_DDC2_FFT_OverlappingBuffer.SetSize(c.m_DDC2_FFT_Size);

    GetNormalizedWindowCoeffs(c.m_DDC2_FFT_Coeffs,c.m_DDC2_FFT_Size,1);

    k=(double)m_DDC2_Info.SampleRate/c.m_DDC2_FFT_Size;

    fFirst=(m_DDC2_Info.SampleRate-m_DDC2_Info.Bandwidth)*0.5/k;                           
    fLast=(m_DDC2_Info.SampleRate-(m_DDC2_Info.SampleRate-m_DDC2_Info.Bandwidth)*0.5)/k;

    First=(int)fFirst;          //the first FFT bin where useful band starts
    Last=(int)ceil(fLast);      //the last FFT bin where useful band ends

    c.m_DDC2_FFT_BinCount=Last-First+1;
    c.m_DDC2_FFT_FirstBin=First;

    c.m_DDC2_FFT_Range=c.m_DDC2_FFT_BinCount*k;
    c.m_DDC2_FFT_Offset=(fFirst-(First+0.5))*k;

    UpdateDDC2SpectrumCompensationData(Channel);

    c.m_DDC2_FFT_Mutex.Unlock();
}

//TReceiver::StartDDC1
//Calculates DDC1 streaming buffer size to be ~UPDATE_INTERVAL ms long (rounded up to 64 samples)
//and starts DDC1 streaming or DDC1 playback.
void TReceiver::StartDDC1(void)
{
    if(m_Power)
    {
        m_DDC1_FFT_OverlappingBuffer.Reset();

        m_DDC1_SamplesPerBuffer=(m_DDC1_Info.SampleRate*UPDATE_INTERVAL/1000+63)&(~63); //~50 ms length buffer, size aligned to 64 samples
        
        m_Device->StartDDC1(m_DDC1_SamplesPerBuffer);
    }
}


//TReceiver::StartDDC2
//Calculates DDC2 streaming buffer size to be ~UPDATE_INTERVAL ms long (rounded up to 64 samples)
//and starts DDC2 streaming.
void TReceiver::StartDDC2(uint32_t Channel)
{
 uint32_t SamplesPerBuffer;    

    if(m_Power)
    {
        m_Channel[Channel].m_DDC2_FFT_OverlappingBuffer.Reset();

        SamplesPerBuffer=(m_DDC2_Info.SampleRate*UPDATE_INTERVAL/1000+63)&(~63); //~50 ms length buffer, size aligned to 64 samples
        m_Device->StartDDC2(Channel,SamplesPerBuffer);
    }
}

//TReceiver::GetDDC1RBWs
//Returns available DDC1 spectrum RBWs in Hz
void TReceiver::GetDDC1RBWs(double *RBWs)
{
 int i;

    for(i=0;i<RBW_COUNT;i++)
    {
        RBWs[i]=(double)m_DDC1_Info.SampleRate/(MIN_FFT_SIZE<<i);
    }
}

//TReceiver::GetDDC2RBWs
//Returns available DDC2 spectrum RBWs in Hz for given channel
void TReceiver::GetDDC2RBWs(double *RBWs)
{
 int i;

    for(i=0;i<RBW_COUNT;i++)
    {
        RBWs[i]=(double)m_DDC2_Info.SampleRate/(MIN_FFT_SIZE<<i);
    }
}

//TReceiver::StartAudio
//Calculates audio streaming buffer size to be ~UPDATE_INTERVAL ms long (rounded up to 64 samples)
//and starts audio streaming or audio playback.
void TReceiver::StartAudio(uint32_t Channel)
{
    if(m_Power)
    {
        m_Channel[Channel].m_Audio_FFT_OverlappingBuffer.Reset();

        m_Channel[Channel].m_Audio_SamplesPerBuffer=(AUDIO_SAMPLE_RATE*UPDATE_INTERVAL/1000+63)&(~63); //~50 ms length buffer, size aligned to 64 samples
        
        m_Device->StartAudio(Channel,m_Channel[Channel].m_Audio_SamplesPerBuffer);        
    }
}

//TReceiver::SetDDC1RBWIndex
//Changes DDC1 RBW
void TReceiver::SetDDC1RBWIndex(uint32_t RBWIndex)
{
    if(m_Ignore)
    {
        return;
    }    

    if(RBWIndex>=RBW_COUNT)
    {
        RBWIndex=RBW_COUNT-1;
    }

    if(m_DDC1_RBW_Index!=RBWIndex)
    {
        m_DDC1_RBW_Index=RBWIndex;
        UpdateDDC1FFTCoeffs();

        InvokeEvent(ReceiverEvent_DDC1_RBWIndex);
    }
}

//TReceiver::SetDDC2RBWIndex
//Changes DDC2 RBW for given channel
void TReceiver::SetDDC2RBWIndex(uint32_t Channel,uint32_t RBWIndex)
{
    if(m_Ignore)
    {
        return;
    }

    if(RBWIndex>=RBW_COUNT)
    {
        RBWIndex=RBW_COUNT-1;
    }

    if(m_Channel[Channel].m_DDC2_RBW_Index!=RBWIndex)
    {
        m_Channel[Channel].m_DDC2_RBW_Index=RBWIndex;
        UpdateDDC2FFTCoeffs(Channel);

        InvokeEvent(Channel,ChannelEvent_DDC2_RBWIndex);
    }
}


//TReceiver::GetAudioRBWs
//Returns available audio spectrum RBWs in Hz for given channel
void TReceiver::GetAudioRBWs(double *RBWs)
{
 int i;

    for(i=0;i<RBW_COUNT;i++)
    {
        RBWs[i]=(double)AUDIO_SAMPLE_RATE/(MIN_FFT_SIZE<<i);
    }
}

//TReceiver::UpdateAudioFFTCoeffs
//Updates audio spectrum FFT size acording to RBW index and normalization/window coeffs
void TReceiver::UpdateAudioFFTCoeffs(uint32_t Channel)
{
    m_Channel[Channel].m_Audio_FFT_Mutex.Lock();

    m_Channel[Channel].m_Audio_FFT_Size=MIN_FFT_SIZE<<m_Channel[Channel].m_Audio_RBW_Index;
    m_Channel[Channel].m_Audio_FFT_OverlappingBuffer.SetSize(m_Channel[Channel].m_Audio_FFT_Size);

    GetNormalizedWindowCoeffs(m_Channel[Channel].m_Audio_FFT_Coeffs,m_Channel[Channel].m_Audio_FFT_Size,1);  

    m_Channel[Channel].m_Audio_FFT_Mutex.Unlock();
}

//TReceiver::SetAudioRBWIndex
//Changes audio spectrum RBW for given channel
void TReceiver::SetAudioRBWIndex(uint32_t Channel,uint32_t RBWIndex)
{
    if(m_Ignore)
    {
        return;
    }

    if(RBWIndex>=RBW_COUNT)
    {
        RBWIndex=RBW_COUNT-1;
    }

    if(m_Channel[Channel].m_Audio_RBW_Index!=RBWIndex)
    {
        m_Channel[Channel].m_Audio_RBW_Index=RBWIndex;
        UpdateAudioFFTCoeffs(Channel);

        InvokeEvent(Channel,ChannelEvent_Audio_RBWIndex);
    }
}

//TReceiver::SetActiveChannel
//Changes active channel (RX1 / RX2 on GUI panel)
void TReceiver::SetActiveChannel(uint32_t Channel)
{
 uint32_t i;

    if(m_Ignore)
    {
        InvokeEvent(ReceiverEvent_ActiveChannel);
        return;
    }

    if(Channel>=m_DeviceInfo.ChannelCount)
    {
        Channel=m_DeviceInfo.ChannelCount-1;
    }

    if(m_ActiveChannel!=Channel)
    {
        m_ActiveChannel=Channel;

        InvokeEvent(ReceiverEvent_ActiveChannel);

        for(i=0;i<m_DeviceInfo.ChannelCount;i++)
        {
            if(m_ActiveChannel==i)
            {
                m_Channel[i].m_Audio_MuteBits&=~MUTE_BITS_INACTIVE_CHANNEL;
            }
            else
            {
                m_Channel[i].m_Audio_MuteBits|=MUTE_BITS_INACTIVE_CHANNEL;
            }

            m_Device->SetMute(i,m_Channel[i].m_Audio_MuteBits!=0);
        }
    }
}


//TReceiver::OnDDC2AbsFrequencyChanged
//Updates DDC2 spectrum compensation data and sends event to GUI when absolute
//DDC2 frequency is changed (e.g. changing DDC1 relative frequency or DDC2 relative frequency)
void TReceiver::OnDDC2AbsFrequencyChanged(uint32_t Channel)
{
    UpdateDDC2SpectrumCompensationData(Channel);

    InvokeEvent(Channel,ChannelEvent_DDC2_AbsFrequency);

    m_Channel[Channel].m_DEM_AbsFrequency=m_Channel[Channel].m_DEM_Frequency+m_Channel[Channel].m_DDC2_AbsFrequency;

    OnDemAbsFrequencyChanged(Channel);
}

//TReceiver::OnDemAbsFrequencyChanged
//Sends event to GUI when absolute demod frequency is changed (e.g. changing DDC1 relative frequency
//or DDC2 relative frequency or demod relative frequency)
void TReceiver::OnDemAbsFrequencyChanged(uint32_t Channel)
{
    InvokeEvent(Channel,ChannelEvent_DEM_AbsFrequency);
}



//TReceiver::SetDemAbsFrequency
//Tune DDC1, DDC2 and demod so that final demod frequency be equal to requested one
void TReceiver::SetDemAbsFrequency(uint32_t Channel,int32_t Frequency)
{
 int32_t i32;

    if(m_Ignore)
    {
        InvokeEvent(Channel,ChannelEvent_DEM_AbsFrequency);
        return;
    }

    if(Frequency==GetDemAbsFrequency(Channel))
    {
        return;
    }

    
    m_Device->SetFrequency(Channel,Frequency);

    m_Device->GetDDC1Frequency((uint32_t*)&i32);

    if(m_DDC1_Frequency!=i32)
    {
        m_DDC1_Frequency=i32;
                
        InvokeEvent(ReceiverEvent_DDC1_Frequency);

        m_Device->GetDemodulatorFrequency(Channel,&i32);

        if(m_Channel[Channel].m_DEM_Frequency!=i32)
        {
            m_Channel[Channel].m_DEM_Frequency=i32;
            InvokeEvent(Channel,ChannelEvent_DEM_Frequency);
        }
        
        m_Device->GetDDC2Frequency(Channel,&i32);

        if(m_Channel[Channel].m_DDC2_Frequency!=i32)
        {
            m_Channel[Channel].m_DDC2_Frequency=i32;
            InvokeEvent(Channel,ChannelEvent_DDC2_Frequency);
        }

        OnDDC1FrequencyChanged(false);
    }
    else
    {
        m_Device->GetDemodulatorFrequency(Channel,&i32);

        if(m_Channel[Channel].m_DEM_Frequency!=i32)
        {
            m_Channel[Channel].m_DEM_Frequency=i32;
            InvokeEvent(Channel,ChannelEvent_DEM_Frequency);
        }
        
        m_Device->GetDDC2Frequency(Channel,&i32);

        if(m_Channel[Channel].m_DDC2_Frequency!=i32)
        {
            m_Channel[Channel].m_DDC2_Frequency=i32;
            InvokeEvent(Channel,ChannelEvent_DDC2_Frequency);
        }

        InvokeEvent(Channel,ChannelEvent_DEM_AbsFrequency);
    }
}

//TReceiver::GetDemBandwidthMinMax
//Returns min and max demod bandwidth which entire falls to DDC2 on current
//relative demod frequency
void TReceiver::GetDemBandwidthMinMax(uint32_t Channel,uint32_t *Min,uint32_t *Max)
{
    *Min=1;

    switch(m_ModeCodes[m_Channel[Channel].m_DEM_Mode])
    {
        case G3XDDC_MODE_LSB:
            *Max=(int32_t)m_DDC2_Info.Bandwidth/2+m_Channel[Channel].m_DEM_Frequency;
            break;

        case G3XDDC_MODE_USB:
            *Max=(int32_t)m_DDC2_Info.Bandwidth/2-m_Channel[Channel].m_DEM_Frequency;
            break;

        default:
            if(m_Channel[Channel].m_DEM_Frequency<0)
            {
                *Max=-(int32_t)m_DDC2_Info.Bandwidth+2*m_Channel[Channel].m_DEM_Frequency;
            }
            else
            {
                *Max=(int32_t)m_DDC2_Info.Bandwidth-2*m_Channel[Channel].m_DEM_Frequency;
            }
            break;
    }
}

//TReceiver::GetDemFrequencyMinMax
//Returns frequency range (min and max relative demod frequency) in which entire demod filter
//falls to DDC2
void TReceiver::GetDemFrequencyMinMax(uint32_t Channel,int32_t *Min,int32_t *Max)
{
    switch(m_ModeCodes[m_Channel[Channel].m_DEM_Mode])
    {
        case G3XDDC_MODE_LSB:
            *Max=m_DDC2_Info.Bandwidth/2;
            *Min=-(int32_t)m_DDC2_Info.Bandwidth/2+(int32_t)m_Channel[Channel].m_DEM_Bandwidth;
            break;

        case G3XDDC_MODE_USB:
            *Min=-(int32_t)m_DDC2_Info.Bandwidth/2;
            *Max=(int32_t)m_DDC2_Info.Bandwidth/2-(int32_t)m_Channel[Channel].m_DEM_Bandwidth;
            break;

        default:
            *Min=(-(int32_t)m_DDC2_Info.Bandwidth+(int32_t)m_Channel[Channel].m_DEM_Bandwidth)/2;
            *Max=((int32_t)m_DDC2_Info.Bandwidth-(int32_t)m_Channel[Channel].m_DEM_Bandwidth)/2;
            break;
    }
}

//TReceiver::OnDemBandwidthChanged
//Update settings and sends event to GUI when demod bandwidth is changed
void TReceiver::OnDemBandwidthChanged(uint32_t Channel)
{
 int32_t Min,Max;

    m_Device->SetDemodulatorFilterBandwidth(Channel,m_Channel[Channel].m_DEM_Bandwidth);

    GetDemFrequencyMinMax(Channel,&Min,&Max);

    InvokeEvent(Channel,ChannelEvent_DEM_Bandwidth);

    if(m_Channel[Channel].m_DEM_Frequency<Min)
    {
        m_Channel[Channel].m_DEM_Frequency=Min;

        OnDemFrequencyChanged(Channel);
    }
    else if(m_Channel[Channel].m_DEM_Frequency>Max)
    {
        m_Channel[Channel].m_DEM_Frequency=Max;

        OnDemFrequencyChanged(Channel);
    }    
}

//TReceiver::OnDemFrequencyChanged
//Update settings and sends event to GUI when demod relative frequency is changed
void TReceiver::OnDemFrequencyChanged(uint32_t Channel)
{
 uint32_t Min,Max;

    m_Device->SetDemodulatorFrequency(Channel,m_Channel[Channel].m_DEM_Frequency);

    InvokeEvent(Channel,ChannelEvent_DEM_Frequency);

    m_Channel[Channel].m_DEM_AbsFrequency=m_Channel[Channel].m_DEM_Frequency+m_Channel[Channel].m_DDC2_AbsFrequency;

    OnDemAbsFrequencyChanged(Channel);

    GetDemBandwidthMinMax(Channel,&Min,&Max);

    if(m_Channel[Channel].m_DEM_Bandwidth<Min)
    {
        m_Channel[Channel].m_DEM_Bandwidth=Min;
        OnDemBandwidthChanged(Channel);
    }
    else if(m_Channel[Channel].m_DEM_Bandwidth>Max)
    {
        m_Channel[Channel].m_DEM_Bandwidth=Max;
        OnDemBandwidthChanged(Channel);
    }
}

//TReceiver::SetDemBandwidth
//Changes demod bandwidth
void TReceiver::SetDemBandwidth(uint32_t Channel,uint32_t Bandwidth)
{
    if(m_Ignore)
    {
        InvokeEvent(Channel,ChannelEvent_DEM_Bandwidth);
        return;
    }

    if(m_Channel[Channel].m_DEM_Bandwidth!=Bandwidth)
    {
        if(Bandwidth<1)
        {
            Bandwidth=1;
        }
        else if(Bandwidth>m_DDC2_Info.Bandwidth)
        {
            Bandwidth=m_DDC2_Info.Bandwidth;
        }

        if(m_Channel[Channel].m_DEM_Bandwidth!=Bandwidth)
        {
            m_Channel[Channel].m_DEM_Bandwidth=Bandwidth;

            OnDemBandwidthChanged(Channel);
        }
        else
        {
            InvokeEvent(Channel,ChannelEvent_DEM_Bandwidth);
        }
    }
}

//TReceiver::OnDDC2FrequencyChanged
//Updates settings and sends event to GUI when relative DDC2 frequency is changed
void TReceiver::OnDDC2FrequencyChanged(uint32_t Channel)
{
    m_Device->SetDDC2Frequency(Channel,m_Channel[Channel].m_DDC2_Frequency);
    
    m_Channel[Channel].m_DDC2_AbsFrequency=m_Channel[Channel].m_DDC2_Frequency+m_DDC1_Frequency;

    InvokeEvent(Channel,ChannelEvent_DDC2_Frequency);

    OnDDC2AbsFrequencyChanged(Channel);
}

//TReceiver::OnDDC1FrequencyChanged
//Updates settings and sends event to GUI when relative DDC1 frequency is changed
void TReceiver::OnDDC1FrequencyChanged(bool UpdateHw)
{
 uint32_t i;

    if(UpdateHw)
    {
        m_Device->SetDDC1Frequency(m_DDC1_Frequency);
    }

    InvokeEvent(ReceiverEvent_DDC1_Frequency);

    UpdateDDC1SpectrumCompensationData();

    for(i=0;i<m_DeviceInfo.ChannelCount;i++)
    {
        m_Channel[i].m_DDC2_AbsFrequency=m_Channel[i].m_DDC2_Frequency+m_DDC1_Frequency;

        OnDDC2AbsFrequencyChanged(i);
    }
}

//TReceiver::SetDDC1Frequency
//Changes relative DDC1 frequency (e.g. dragging the band inside IF spectrum)
void TReceiver::SetDDC1Frequency(int32_t Frequency)
{
 int32_t Min,Max;

    if(m_Ignore)
    {
        InvokeEvent(ReceiverEvent_DDC1_Frequency);
        return;
    }

    if(m_DDC1_Frequency!=Frequency)
    {
        Min=m_DDC1_Info.Bandwidth/2;
        Max=50000000-m_DDC1_Info.Bandwidth/2;

        if(Frequency<Min)
        {
            Frequency=Min;
        }
        else if(Frequency>Max)
        {
            Frequency=Max;
        }

        if(m_DDC1_Frequency!=Frequency)
        {
            m_DDC1_Frequency=Frequency;

            OnDDC1FrequencyChanged(true);
        }
        else
        {
            InvokeEvent(ReceiverEvent_DDC1_Frequency);
        }
    }
}

//TReceiver::SetDDC2Frequency
//Changes relative DDC2 frequency (e.g. dragging the band inside DDC1 spectrum)
void TReceiver::SetDDC2Frequency(uint32_t Channel,int32_t Frequency)
{
 int32_t Min,Max;

    if(m_Ignore)
    {
        InvokeEvent(Channel,ChannelEvent_DDC2_Frequency);
        return;
    }

    if(m_Channel[Channel].m_DDC2_Frequency!=Frequency)
    {
        Max=(m_DDC1_Info.Bandwidth-m_DDC2_Info.Bandwidth)/2;
        Min=-Max;

        if(Frequency<Min)
        {
            Frequency=Min;
        }
        else if(Frequency>Max)
        {
            Frequency=Max;
        }

        if(m_Channel[Channel].m_DDC2_Frequency!=Frequency)
        {
            m_Channel[Channel].m_DDC2_Frequency=Frequency;

            OnDDC2FrequencyChanged(Channel);
        }
        else
        {
            InvokeEvent(Channel,ChannelEvent_DDC2_Frequency);
        }
    }
}

//TReceiver::SetDemFrequency
//Changes relative demod filter frequency (e.g. dragging the band inside DDC2 spectrum)
void TReceiver::SetDemFrequency(uint32_t Channel,int32_t Frequency)
{
 int32_t Min,Max;

    if(m_Ignore)
    {
        InvokeEvent(Channel,ChannelEvent_DEM_Frequency);
        return;
    }

    if(m_Channel[Channel].m_DEM_Frequency!=Frequency)
    {
        GetDemFrequencyMinMax(Channel,&Min,&Max);

        if(Frequency<Min)
        {
            Frequency=Min;
        }
        else if(Frequency>Max)
        {
            Frequency=Max;
        }

        if(m_Channel[Channel].m_DEM_Frequency!=Frequency)
        {
            m_Channel[Channel].m_DEM_Frequency=Frequency;

            OnDemFrequencyChanged(Channel);
        }
        else
        {
            InvokeEvent(Channel,ChannelEvent_DEM_Frequency);
        }
    }
}



//TReceiver::GetDemModeName
//Returns mode name for given mode index
wxString TReceiver::GetDemModeName(uint32_t ModeIndex)
{
    return m_ModeNames[ModeIndex];
}


//TReceiver::SetDemMode
//Changes current demodulator mode
void TReceiver::SetDemMode(uint32_t Channel,uint32_t ModeIndex)
{
    if(m_Ignore)
    {
        InvokeEvent(Channel,ChannelEvent_DEM_Mode);
        return;
    }

    if(m_Channel[Channel].m_DEM_Mode!=ModeIndex)
    {
        m_Channel[Channel].m_DEM_Mode=ModeIndex;

        OnDemModeChanged(Channel);
    }
}

//TReceiver::OnDemModeChanged
//Updates settings and sends events when demodulator mode is changed
void TReceiver::OnDemModeChanged(uint32_t Channel)
{
 uint32_t Min,Max;     

    GetDemBandwidthMinMax(Channel,&Min,&Max);

    if(m_Channel[Channel].m_DEM_Bandwidth<Min)
    {
        m_Channel[Channel].m_DEM_Bandwidth=Min;
        OnDemBandwidthChanged(Channel);
    }
    else if(m_Channel[Channel].m_DEM_Bandwidth>Max)
    {
        m_Channel[Channel].m_DEM_Bandwidth=Max;
        OnDemBandwidthChanged(Channel);
    }
  

    m_Device->SetDemodulatorMode(Channel,m_ModeCodes[m_Channel[Channel].m_DEM_Mode]);

    InvokeEvent(Channel,ChannelEvent_DEM_Mode);
}

    
//TReceiver::UpdateIFSpectrumCompensationData
//Updates IF spectrum compensation data
void TReceiver::UpdateIFSpectrumCompensationData(void)
{
    m_IF_FFT_Mutex.Lock();

    m_Device->GetSpectrumCompensation(50000000/2,50000000,m_IF_SpectrumCompensation,m_IF_FFT_Size/2);

    m_IF_FFT_Mutex.Unlock();
}

//TReceiver::UpdateDDC1SpectrumCompensationData
//Updates DDC1 spectrum compensation data
void TReceiver::UpdateDDC1SpectrumCompensationData(void)
{
    m_DDC1_FFT_Mutex.Lock();

    m_Device->GetSpectrumCompensation(m_DDC1_Frequency,m_DDC1_Info.SampleRate,m_DDC1_SpectrumCompensation,m_DDC1_FFT_Size);

    m_DDC1_FFT_Mutex.Unlock();
}

//TReceiver::UpdateDDC2SpectrumCompensationData
//Updates DDC2 spectrum compensation data
void TReceiver::UpdateDDC2SpectrumCompensationData(uint32_t Channel)
{
    m_Channel[Channel].m_DDC2_FFT_Mutex.Lock();

    m_Device->GetSpectrumCompensation(m_Channel[Channel].m_DDC2_AbsFrequency,m_DDC2_Info.SampleRate,
        m_Channel[Channel].m_DDC2_SpectrumCompensation,
        m_Channel[Channel].m_DDC2_FFT_Size);

    m_Channel[Channel].m_DDC2_FFT_Mutex.Unlock();
}

//TReceiver::UpdateSpectrumCompensationData
//Updates compensation data for all the spectrums
void TReceiver::UpdateSpectrumCompensationData(void)
{
 uint32_t i;

    UpdateIFSpectrumCompensationData();
    UpdateDDC1SpectrumCompensationData();

    for(i=0;i<m_DeviceInfo.ChannelCount;i++)
    {        
        UpdateDDC2SpectrumCompensationData(i);
    }
}

//TReceiver::GetPowerDirect
//Returns receiver's power state directly read from the API
bool TReceiver::GetPowerDirect(void)
{
 BOOL Power;
 bool b;

    m_Device->GetPower(&Power);

    b=Power!=FALSE;

    if(m_Power!=b)
    {
        m_Power=b;
        OnPowerChanged();
    }

    return b;
}


//TReceiver::UpdateAGC
//Updates AGC params according to AGC type
void TReceiver::UpdateAGC(uint32_t Channel)
{
    switch(m_Channel[Channel].m_AGC)
    {
        case AGC_OFF:
            m_Device->SetAGC(Channel,FALSE);
            break;

        case AGC_SLOW:
            m_Device->SetAGCParams(Channel,0.025,4.0,-15);
            m_Device->SetAGC(Channel,TRUE);
            break;

        case AGC_MEDIUM:
            m_Device->SetAGCParams(Channel,0.015,2.0,-15);
            m_Device->SetAGC(Channel,TRUE);
            break;

        case AGC_FAST:
            m_Device->SetAGCParams(Channel,0.005,0.2,-15);
            m_Device->SetAGC(Channel,TRUE);
            break;
    }
}

//TReceiver::SetAGC
//Changes AGC
void TReceiver::SetAGC(uint32_t Channel,uint32_t AGC)
{
    if(m_Ignore)
    {
        InvokeEvent(Channel,ChannelEvent_AGC);
        return;
    }

    if(m_Channel[Channel].m_AGC!=AGC)
    {
        m_Channel[Channel].m_AGC=AGC;

        UpdateAGC(Channel);
    }

    InvokeEvent(Channel,ChannelEvent_AGC);
}

//TReceiver::SetGain
//Changes gain of signal before demodulation
void TReceiver::SetGain(uint32_t Channel,int32_t Gain)
{
    if(!m_Ignore)
    {
        if(m_Channel[Channel].m_Gain!=Gain)
        {
            m_Channel[Channel].m_Gain=Gain;
            m_Device->SetGain(Channel,Gain);
        }
    }    

    InvokeEvent(Channel,ChannelEvent_Gain);
}

//TReceiver::SetAudioFilter
//Enables/disables audio filter
void TReceiver::SetAudioFilter(uint32_t Channel,bool Enabled)
{
    if(!m_Ignore)
    {
        if(m_Channel[Channel].m_Audio_Filter!=Enabled)
        {
            m_Channel[Channel].m_Audio_Filter=Enabled;

            m_Device->SetAudioFilter(Channel,Enabled);
        }
    }

    InvokeEvent(Channel,ChannelEvent_Audio_Filter);
}

//TReceiver::SetAudioFilterParams
//Changes audio filter params
void TReceiver::SetAudioFilterParams(uint32_t Channel,uint32_t Low,uint32_t High,double Deem)
{
    if(!m_Ignore)
    {
        if(m_Channel[Channel].m_Audio_Filter_CutOffLow!=Low ||
            m_Channel[Channel].m_Audio_Filter_CutOffHigh!=High ||
            m_Channel[Channel].m_Audio_Filter_Deemphasis!=Deem)
        {
            m_Device->SetAudioFilterParams(Channel,Low,High,Deem);
            m_Device->GetAudioFilterParams(Channel,&m_Channel[Channel].m_Audio_Filter_CutOffLow,
                &m_Channel[Channel].m_Audio_Filter_CutOffHigh,
                &m_Channel[Channel].m_Audio_Filter_Deemphasis);
        }
    }

    InvokeEvent(Channel,ChannelEvent_Audio_FilterParams);
}

//TReceiver::SetDemAMSSideband
//Changes side band (upper/lower/both) used for AMS mode
void TReceiver::SetDemAMSSideband(uint32_t Channel,uint32_t Sideband)
{
    if(m_Channel[Channel].m_DEM_AMSSideband!=Sideband)
    {
        if(!m_Ignore)
        {
            m_Device->SetDemodulatorParam(Channel,G3XDDC_DEMODULATOR_PARAM_AMS_SIDE_BAND,&Sideband,sizeof(Sideband));
            m_Device->GetDemodulatorParam(Channel,G3XDDC_DEMODULATOR_PARAM_AMS_SIDE_BAND,&m_Channel[Channel].m_DEM_AMSSideband,sizeof(m_Channel[Channel].m_DEM_AMSSideband));
        }

        InvokeEvent(Channel,ChannelEvent_DEM_AMSSideband);
    }
}

//TReceiver::SetDemDSBSideband
//Changes side band (upper/lower/both) used for DSB mode
void TReceiver::SetDemDSBSideband(uint32_t Channel,uint32_t Sideband)
{    
    if(m_Channel[Channel].m_DEM_DSBSideband!=Sideband)
    {
        if(!m_Ignore)
        {
            m_Device->SetDemodulatorParam(Channel,G3XDDC_DEMODULATOR_PARAM_DSB_SIDE_BAND,&Sideband,sizeof(Sideband));
            m_Device->GetDemodulatorParam(Channel,G3XDDC_DEMODULATOR_PARAM_DSB_SIDE_BAND,&m_Channel[Channel].m_DEM_DSBSideband,sizeof(m_Channel[Channel].m_DEM_DSBSideband));
        }

        InvokeEvent(Channel,ChannelEvent_DEM_DSBSideband);
    }
}

//TReceiver::SetDemISBSideband
//Changes side band (upper/lower/both) used for ISB mode
void TReceiver::SetDemISBSideband(uint32_t Channel,uint32_t Sideband)
{
    if(m_Channel[Channel].m_DEM_ISBSideband!=Sideband)
    {
        if(!m_Ignore)
        {
            m_Device->SetDemodulatorParam(Channel,G3XDDC_DEMODULATOR_PARAM_ISB_SIDE_BAND,&Sideband,sizeof(Sideband));
            m_Device->GetDemodulatorParam(Channel,G3XDDC_DEMODULATOR_PARAM_ISB_SIDE_BAND,&m_Channel[Channel].m_DEM_ISBSideband,sizeof(m_Channel[Channel].m_DEM_ISBSideband));
        }

        InvokeEvent(Channel,ChannelEvent_DEM_ISBSideband);
    }
}

//TReceiver::SetDemCWTone
//Changes CW tone frequency
void TReceiver::SetDemCWTone(uint32_t Channel,uint32_t Frequency)
{
    if(m_Channel[Channel].m_DEM_CWTone!=Frequency)
    {
        if(!m_Ignore)
        {
            m_Device->SetDemodulatorParam(Channel,G3XDDC_DEMODULATOR_PARAM_CW_FREQUENCY,&Frequency,sizeof(Frequency));
            m_Device->GetDemodulatorParam(Channel,G3XDDC_DEMODULATOR_PARAM_CW_FREQUENCY,&m_Channel[Channel].m_DEM_CWTone,sizeof(m_Channel[Channel].m_DEM_CWTone));
        }

        InvokeEvent(Channel,ChannelEvent_DEM_CWTone);
    }
}



//TReceiver::StopDDC1
//Stops DDC1, DDC2 and audio streaming (it does not stop audio playback if active)
void TReceiver::StopDDC1(void)
{
    m_Device->StopDDC1();
}

//TReceiver::StopDDC2
//Stops DDC2 and audio streaming (it does not stop audio playback if active)
void TReceiver::StopDDC2(uint32_t Channel)
{
    m_Device->StopDDC2(Channel);
}

//TReceiver::StopAudio
//Stops audio streaming (it stops audio playback too)
void TReceiver::StopAudio(uint32_t Channel)
{
    m_Device->StopAudio(Channel);
}

//TReceiver::SetPreselectors
//Sets preselectors
void TReceiver::SetPreselectors(uint32_t CutOffLowIndex,uint32_t CutOffHighIndex)
{
    if(m_Ignore)
    {
        return;
    }

    if(CutOffLowIndex>=PRESEL_COUNT)
    {
        CutOffLowIndex=PRESEL_COUNT-1;
    }

    if(CutOffHighIndex>=PRESEL_COUNT)
    {
        CutOffHighIndex=PRESEL_COUNT-1;
    }

    while(m_Presel_CutOffLow[CutOffLowIndex]>m_Presel_CutOffHigh[CutOffHighIndex])
    {
        CutOffHighIndex++;
    }

    m_Device->SetPreselectors(m_Presel_CutOffLow[CutOffLowIndex],m_Presel_CutOffHigh[CutOffHighIndex]);

    m_Presel_CutOffLowIndex=CutOffLowIndex;
    m_Presel_CutOffHighIndex=CutOffHighIndex;

    InvokeEvent(ReceiverEvent_Preselectors);
}

//TReceiver::UpdateAudioVolumeCoeff
//Converts volume value (0 - 31) to a value which is multiplied with audio samples
void TReceiver::UpdateAudioVolumeCoeff(uint32_t Channel)
{
    m_Channel[Channel].m_Audio_VolumeCoeff=(float)(m_Channel[Channel].m_Audio_Volume/31.0);
}

//TReceiver::SetVolume
//Sets audio volume
void TReceiver::SetVolume(uint32_t Channel,uint8_t Volume)
{
    if(m_Ignore)
    {
        InvokeEvent(Channel,ChannelEvent_Audio_Volume);
        return;
    }

    m_Device->SetVolume(Channel,Volume);

    InvokeEvent(Channel,ChannelEvent_Audio_Volume);
}

//TReceiver::GetVolume
//Returns current audio volume
uint8_t TReceiver::GetVolume(uint32_t Channel)
{
 uint8_t Volume;

    m_Device->GetVolume(Channel,&Volume);

    return Volume;
}

//TReceiver::SetMute
//Mutes/unmutes audio
void TReceiver::SetMute(uint32_t Channel,bool Mute)
{
    if(Mute)
    {
        m_Channel[Channel].m_Audio_MuteBits|=MUTE_BITS_MANUAL;
    }
    else
    {
        m_Channel[Channel].m_Audio_MuteBits&=~MUTE_BITS_MANUAL;
    }

    m_Device->SetMute(Channel,m_Channel[Channel].m_Audio_MuteBits!=0);

    InvokeEvent(Channel,ChannelEvent_Audio_Mute);
}

//TReceiver::SetSquelched
//Mutes/unmutes audio acording to the Squelched param
void TReceiver::SetSquelched(uint32_t Channel,bool Squelched)
{
    if(Squelched)
    {
        m_Channel[Channel].m_Audio_MuteBits|=MUTE_BITS_SQUELCH;
    }
    else
    {
        m_Channel[Channel].m_Audio_MuteBits&=~MUTE_BITS_SQUELCH;
    }

    m_Device->SetMute(Channel,m_Channel[Channel].m_Audio_MuteBits!=0);

    InvokeEvent(Channel,ChannelEvent_Audio_Mute);
}