/*****************************************************************************/
/*                      WiNRADiO G35DDC EXCALIBUR ULTRA                      */
/*                                SDK DEMO                                   */
/*             Simple playback on waveform-audio output device               */
/*                        (C) Copyright RADIXON 2014                         */
/*****************************************************************************/

#include "u_waveout.h"
#include <mmreg.h>
#include <process.h>

TWaveOut::TWaveOut(void)
{
    m_hWaveOut=NULL;
    InitializeCriticalSection(&m_CS);

    m_hThread=(HANDLE)_beginthreadex(NULL,0,StaticThreadProcedure,this,0,&m_ThreadId);
}

TWaveOut::~TWaveOut(void)
{
    Close();

    PostThreadMessage(m_ThreadId,WM_QUIT,0,0);

    WaitForSingleObject(m_hThread,INFINITE);

    CloseHandle(m_hThread);

    DeleteCriticalSection(&m_CS);
}

void TWaveOut::Close(void)
{
 HWAVEOUT hWaveOut;
 int i;

    if(m_hWaveOut==NULL)
    {
        return;
    }

    hWaveOut=m_hWaveOut;

    EnterCriticalSection(&m_CS);

    m_hWaveOut=NULL;

    LeaveCriticalSection(&m_CS);

    waveOutReset(hWaveOut);

    while(m_PlayCount>0)
    {
        Sleep(1);
    }

    for(i=0;i<WAVEOUT_BUFFER_COUNT;i++)
    {
        waveOutUnprepareHeader(hWaveOut,m_Headers+i,sizeof(m_Headers[i]));
    }

    waveOutClose(hWaveOut);

    free(m_Buffers[0]);

    m_Avail.clear();
}

UINT __stdcall TWaveOut::StaticThreadProcedure(void *This)
{
    ((TWaveOut*)This)->ThreadProcedure();
    return 0;
}

void TWaveOut::ThreadProcedure(void)
{
 MSG Msg;
 WAVEHDR *Header;

    while(GetMessage(&Msg,NULL,0,0))
    {
        if(Msg.message==MM_WOM_DONE)
        {
            Header=(WAVEHDR*)Msg.lParam;

            Header->dwFlags=WHDR_PREPARED;

            EnterCriticalSection(&m_CS);

            m_Avail.push_back(Header);
            InterlockedDecrement(&m_PlayCount);

            if(!m_Buffering && m_PlayCount==0)
            {
                m_Buffering=true;
                waveOutPause(m_hWaveOut);
            }

            LeaveCriticalSection(&m_CS);
        }
    }
}

bool TWaveOut::Open(uint32_t SampleRate)
{
 WAVEFORMATEX Format;
 int i;
 HWAVEOUT hWaveOut;

    Close();

    Format.wFormatTag=WAVE_FORMAT_IEEE_FLOAT;
    Format.cbSize=0;
    Format.nChannels=1;
    Format.nSamplesPerSec=SampleRate;
    Format.wBitsPerSample=32;
    Format.nBlockAlign=(Format.nChannels*Format.wBitsPerSample)/8;
    Format.nAvgBytesPerSec=Format.nSamplesPerSec*Format.nBlockAlign;

    if(waveOutOpen(&hWaveOut,WAVE_MAPPER,&Format,m_ThreadId,0,CALLBACK_THREAD)!=MMSYSERR_NOERROR)
    {
        Format.wFormatTag=WAVE_FORMAT_PCM;
        Format.wBitsPerSample=16;
        Format.nBlockAlign=(Format.nChannels*Format.wBitsPerSample)/8;
        Format.nAvgBytesPerSec=Format.nSamplesPerSec*Format.nBlockAlign;

        if(waveOutOpen(&hWaveOut,WAVE_MAPPER,&Format,m_ThreadId,0,CALLBACK_THREAD)!=MMSYSERR_NOERROR)
        {
            return false;
        }

        m_Float=false;
    }
    else
    {
        m_Float=true;
    }

    m_SamplesPerBuffer=(SampleRate*50/1000+63)&(~63); //50 ms buffer length, aligned to 64 sample sets
    m_BufferSize=m_SamplesPerBuffer*Format.nBlockAlign;

    m_Buffers[0]=(char*)malloc(WAVEOUT_BUFFER_COUNT*m_BufferSize);

    if(m_Buffers[0]==NULL)
    {
        waveOutClose(hWaveOut);
        return false;
    }

    for(i=1;i<WAVEOUT_BUFFER_COUNT;i++)
    {
        m_Buffers[i]=m_Buffers[i-1]+m_BufferSize;
    }

    ZeroMemory(m_Headers,sizeof(m_Headers));

    for(i=0;i<WAVEOUT_BUFFER_COUNT;i++)
    {
        m_Headers[i].dwBufferLength=m_BufferSize;
        m_Headers[i].lpData=m_Buffers[i];

        if(waveOutPrepareHeader(hWaveOut,m_Headers+i,sizeof(m_Headers[i]))!=MMSYSERR_NOERROR)
        {
            while(!m_Avail.empty())
            {
                waveOutUnprepareHeader(hWaveOut,m_Avail.front(),sizeof(WAVEHDR));
                m_Avail.pop_front();
            }            

            waveOutClose(hWaveOut);

            free(m_Buffers[0]);

            return false;
        }

        m_Avail.push_back(m_Headers+i);
    }

    m_WorkingHeader=NULL;
    m_PlayCount=0;

    waveOutPause(hWaveOut);
    m_Buffering=true;

    EnterCriticalSection(&m_CS);

    m_hWaveOut=hWaveOut;

    LeaveCriticalSection(&m_CS);

    return true;
}

void TWaveOut::Write(const float *Samples,uint32_t Count)
{
 uint32_t Remain;
 uint32_t i;
 short *Buffer;

    EnterCriticalSection(&m_CS);

    if(m_hWaveOut!=NULL)
    {
        if(m_Float)
        {
            while(Count)
            {
                if(m_WorkingHeader==NULL)
                {
                    if(m_Avail.empty())
                    {
                        //no available buffer
                        break;
                    }

                    m_WorkingHeader=m_Avail.back();
                    m_Avail.pop_back();                    
                    m_WorkingPos=0;
                }

                Remain=m_SamplesPerBuffer-m_WorkingPos;
 
                if(Remain<=Count)
                {
                    memcpy(m_WorkingHeader->lpData+m_WorkingPos*sizeof(float),Samples,Remain*sizeof(float));

                    if(waveOutWrite(m_hWaveOut,m_WorkingHeader,sizeof(*m_WorkingHeader))!=MMSYSERR_NOERROR)
                    {
                        break;
                    }                    

                    InterlockedIncrement(&m_PlayCount);

                    if(m_Buffering && m_PlayCount>=3)
                    {
                        waveOutRestart(m_hWaveOut);
                        m_Buffering=false;
                    }

                    Count-=Remain;
                    Samples+=Remain;
                    m_WorkingHeader=NULL;
                }
                else
                {
                    memcpy(m_WorkingHeader->lpData+m_WorkingPos*sizeof(float),Samples,Count*sizeof(float));

                    m_WorkingPos+=Count;
                    break;
                }
            }
        }
        else
        {
            while(Count)
            {
                if(m_WorkingHeader==NULL)
                {
                    if(m_Avail.empty())
                    {
                        //no available buffer
                        break;
                    }

                    m_WorkingHeader=m_Avail.back();
                    m_Avail.pop_back();                    
                    m_WorkingPos=0;
                }

                Remain=m_SamplesPerBuffer-m_WorkingPos;

                Buffer=(short*)m_WorkingHeader->lpData+m_WorkingPos;

                if(Remain<=Count)
                {                    
                    for(i=0;i<Remain;i++)
                    {
                        Buffer[i]=(short)(Samples[i]*32767);
                    }                    

                    if(waveOutWrite(m_hWaveOut,m_WorkingHeader,sizeof(*m_WorkingHeader))!=MMSYSERR_NOERROR)
                    {
                        break;
                    }

                    InterlockedIncrement(&m_PlayCount);

                    if(m_Buffering && m_PlayCount>=3)
                    {
                        waveOutRestart(m_hWaveOut);
                        m_Buffering=false;
                    }

                    Count-=Remain;
                    Samples+=Remain;
                    m_WorkingHeader=NULL;
                }
                else
                {
                    for(i=0;i<Remain;i++)
                    {
                        Buffer[i]=(short)(Samples[i]*32767);
                    }                    

                    m_WorkingPos+=Count;
                    break;
                }
            }
        }
    }

    LeaveCriticalSection(&m_CS);
}
